// Copyright 2019-2024 Yuji Tanaka
// HP : http://galaxystar.image.coocan.jp/
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
/*
 * Othello
 * 
 * 2024' 05/22 Yuji Tanaka
 */
#include <SPI.h>
#include <ILI9163_buf.h>
#include "pitches.h"

#define TFT_DC 48
#define TFT_CS 47
#define TFT_RST 45

ILI9163_buf tft = ILI9163_buf(TFT_CS, TFT_DC, TFT_RST);

int _width, _height;

int StickXPin = 6;
int StickYPin = 7;
int StickX;
int StickY;

int ButtonAPin = 15;
int ButtonBPin = 16;
int ButtonA;
int ButtonB;

int SoundPin = 17;
static int tone_init = 0;

#define XMAX 128
#define YMAX 160
#define SIZE 15

#define random(a,b) ((rand() % ((b)-(a)+1))+(a))

static int state;
static int mode;
static int ban;
static int sente;
static int komasuu;
static int latest_x, latest_y;
char board[8][8];

int mx, my, mchange;
int mx0, mx1, my0, my1;

void setup()
{
    Serial.begin(115200);

    tft.begin(40000000);
    tft.setRotation(2);
    tft.clearDisplay();
    _width=tft.width();
    _height=tft.height();
    tft.display();

    pinMode(ButtonAPin, INPUT_PULLUP);
    pinMode(ButtonBPin, INPUT_PULLUP);
  
  state=0;
}

void loop(){
  int x, y, z;
  uint8_t reg[80];

  StickX = analogRead(StickXPin)/4;
  StickY = 1023 - analogRead(StickYPin)/4;
  ButtonA = digitalRead(ButtonAPin);
  ButtonB = digitalRead(ButtonBPin);

    if (state==0){
      mx0=0;
      my0=0;
      mx1=0;
      my1=4;
      showTitle(my);
      state=1;
    }else if (state==2){
      mx0=0;
      my0=0;
      mx1=0;
      my1=1;
      showSente(my);
      state=3;
    }else if (state==4){
      mx0=0;
      my0=0;
      mx1=7;
      my1=7;
      for(y=0;y<8;y++)
        for(x=0;x<8;x++)
          board[y][x]=0;
      board[3][3]=1;
      board[4][4]=1;
      board[3][4]=2;
      board[4][3]=2;
      ban=1;
      komasuu=4;
      latest_x=-1;
      latest_y=-1;
      show_board((char *)board, mode, mx, my);
      state=5;
    }else if (state==6){
      showEnd();
      state=7;
    }
    if (state==5 && mode>0){
      mx0=0;
      my0=0;
      mx1=7;
      my1=7;
      // Computer
      if ((sente==1 && ban==2) || (sente==0 && ban==1)){
        think((char *)board, &x, &y, ban, mode);
        if (put_on_board((char *)board, x, y, ban)){
          latest_x=x;
          latest_y=y;
          komasuu++;
          if (ban==1)
            ban=2;
          else
            ban=1;
          show_board((char *)board, mode, mx, my);
        }
        // Cannot put
        if (put_num((char *)board, ban)==0){
          if (ban==1)
            ban=2;
          else
            ban=1;
          // Finish
          if (put_num((char *)board, ban)==0)
            state=6;
        }
        return;
      }
    }

    mchange=1;
    if (StickX<256)
      mx--;
    else if (StickX>768)
      mx++;
    else if (StickY<256)
      my--;
    else if (StickY>768)
      my++;
    else
      mchange=0;
    if (mx<mx0) mx=mx0;
    if (mx>mx1) mx=mx1;
    if (my<my0) my=my0;
    if (my>my1) my=my1;

    if (mchange==1){
      if (state==1) state=0;
      if (state==3) state=2;
      if (state==5){
        show_board((char *)board, mode, mx, my);
      }
    }

    if (ButtonA==1){
      delay(100);
      return;
    }

    if (state==1){
      y=my+3;
      tft.setCursor(100, 0);
      if (y==3)
        mode=0;
      if (y==4)
        mode=1;
      if (y==5)
        mode=2;
      if (y==6)
        mode=3;
      if (y==7)
        mode=4;
      if (y==3)
        state=4;
      else if (4<=y && y<=7){
        my=0;
        state=2;
      }
    }else if (state==3){
      y=my+3;
      if (y==3)
        sente=1;
      else if (y==4)
        sente=0;
      if (y==3 || y==4){
        mx=0;
        my=0;
        state=4;
      }
    }else if (state==5){
      x=mx;
      y=my;
      if (0<=x && x<8 && 0<=y && y<8){
        if (put_on_board((char *)board, x, y, ban)){
          latest_x=x;
          latest_y=y;
          komasuu++;

          if (ban==1)
            ban=2;
          else
            ban=1;
        }
        // Cannot put
        if (put_num((char *)board, ban)==0){
          if (ban==1)
            ban=2;
          else
            ban=1;
          // Finish
          if (put_num((char *)board, ban)==0)
            state=6;
        }
        show_board((char *)board, mode, mx, my);
      }
    }else if (state==7){
      state=0;
    }

    do{
      ButtonA = digitalRead(ButtonAPin);
    }while(ButtonA==0);

    delay(200);
}

static void showTitle(int sel)
{
  tft.fillRect(0, 0, _width, _height, ILI9163_BLACK);
  tft.fillRect(0, 0, _width, _height, ILI9163_BLACK);
  tft.setTextSize(3);
  tft.setTextColor(ILI9163_GREEN);
  tft.setCursor(_width/2-64, _height/8);
  tft.print("Othello");

  tft.fillRect(8, _height*(3+sel)/8, 112, 8, ILI9163_RED);

  tft.setTextSize(1);
  tft.setTextColor(ILI9163_WHITE);
  tft.setCursor(8, _height*3/8);
  tft.print("Human vs Human");
  tft.setCursor(8, _height*4/8);
  tft.print("Human vs Computer1");
  tft.setCursor(8, _height*5/8);
  tft.print("Human vs Computer2");
  tft.setCursor(8, _height*6/8);
  tft.print("Human vs Computer3");
  tft.setCursor(8, _height*7/8);
  tft.print("Human vs Computer4");
  tft.display();
}

static void showSente(int sel)
{
  tft.fillRect(0, 0, _width, _height, ILI9163_BLACK);
  tft.fillRect(0, 0, _width, _height, ILI9163_BLACK);
  tft.setTextSize(1);
  tft.setTextColor(ILI9163_GREEN);
  tft.setCursor(0, _height/8);
  tft.print("Select order");

  tft.fillRect(32, _height*(3+sel)/8, 48, 8, ILI9163_RED);

  tft.setTextSize(1);
  tft.setTextColor(ILI9163_WHITE);
  tft.setCursor(32, _height*3/8);
  tft.print("1.First");
  tft.setCursor(32, _height*4/8);
  tft.print("2.Second");
  tft.display();
}

static void showEnd()
{
  show_board((char *)board, mode+8, -1, -1);
}

static void show_board(char *bd, int mm, int cx, int cy)
{
  int x, y;
  int sx, sy;
  char str[32];
  sx=0;
  sy=_height-1-SIZE*8;

  tft.fillRect(0, 0, _width, _height, ILI9163_DARKGREY);
  tft.fillRect(0, 0, _width, _height, ILI9163_DARKGREY);
  tft.fillRect(sx, sy, 8*SIZE, 8*SIZE, ILI9163_GREEN);
  if (latest_x>=0 && latest_y>=0)
    tft.fillRect(sx+latest_x*SIZE+1, sy+latest_y*SIZE+1, SIZE-1, SIZE-1, ILI9163_OLIVE);
  for(x=0;x<=8;x++){
    tft.drawLine(sx+x*SIZE, sy, sx+x*SIZE, sy+8*SIZE, ILI9163_WHITE);
    tft.drawLine(sx, sy+x*SIZE, sx+8*SIZE, sy+x*SIZE, ILI9163_WHITE);
  }
  for(y=0;y<8;y++){
    for(x=0;x<8;x++){
      if (bd[y*8+x]==1)
        tft.fillCircle(sx+x*SIZE+SIZE/2, sy+y*SIZE+SIZE/2, SIZE/2-2, ILI9163_BLACK);
      else if (bd[y*8+x]==2)
        tft.fillCircle(sx+x*SIZE+SIZE/2, sy+y*SIZE+SIZE/2, SIZE/2-2, ILI9163_WHITE);
    }
  }
  if (cx>=0 && cy>=0)
    tft.drawRect(sx+cx*SIZE, sy+cy*SIZE, SIZE+1, SIZE+1, ILI9163_RED);

  tft.setTextSize(1);
  tft.setTextColor(ILI9163_RED);
  tft.setCursor(0, 0);
  if (mm % 8==0)
    tft.print("Human vs Human");
  if (mm % 8==1)
    if (sente==1)
      tft.print("Human vs Computer1");
    else
      tft.print("Computer1 vs Human");
  if (mm % 8==2)
    if (sente==1)
      tft.print("Human vs Computer2");
    else
      tft.print("Computer2 vs Human");
  if (mm % 8==3)
    if (sente==1)
      tft.print("Human vs Computer3");
    else
      tft.print("Computer3 vs Human");
  if (mm % 8==4)
    if (sente==1)
      tft.print("Human vs Computer4");
    else
      tft.print("Computer4 vs Human");

  tft.fillCircle(SIZE*1, SIZE*1, SIZE/2-2, ILI9163_BLACK);
  tft.fillCircle(SIZE*1, SIZE*2, SIZE/2-2, ILI9163_WHITE);
  tft.setTextSize(1);
  tft.setTextColor(ILI9163_WHITE);
  tft.setCursor(SIZE*3-4, SIZE*1-4);
  sprintf(str, "%d", count_point(bd, 1));
  tft.print(str);
  tft.setCursor(SIZE*3-4, SIZE*2-4);
  sprintf(str, "%d", count_point(bd, 2));
  tft.print(str);

  if (mm<8){
    if (ban==1)
      tft.fillCircle(SIZE*9,SIZE*7, SIZE/2-2, ILI9163_BLACK);
    else
      tft.fillCircle(SIZE*9,SIZE*7, SIZE/2-2, ILI9163_WHITE);

    tft.setTextSize(1);
    tft.setTextColor(ILI9163_BLUE);
    tft.setCursor(SIZE*4, SIZE*2-4);
    tft.print("Next");
  }else{
    tft.setTextColor(ILI9163_BLUE);
    tft.setCursor(SIZE*4, SIZE*2-4);
    tft.print("Finish");
  }
  tft.display();
}

static int count_point(char *bd, int c)
{
  int rx, ry;
  int num;
  num=0;
  for(ry=0;ry<8;ry++)
    for(rx=0;rx<8;rx++)
      if(bd[ry*8+rx]==c)
        num++;
  return num;
}

static bool can_put(char *bd,int x,int y,int c)
{
  int rx,ry,rc;
  bool do_rev;
  do_rev=false;
  if(bd[y*8+x]!=0)
          return(do_rev);
  if(c==1) rc=2;
        else  rc=1;
// up
  rx=x;
  ry=y-1;
    for(;;){
    if(ry<0 || bd[ry*8+rx]!=rc)
      break;
    ry--;
    if(ry>=0 && bd[ry*8+rx]==c)
      return true;
  }
// down
  rx=x;
  ry=y+1;
    for(;;){
    if(ry>=8 || bd[ry*8+rx]!=rc)
      break;
    ry++;
    if(ry<8 && bd[ry*8+rx]==c)
      return true;
  }
// left
  rx=x-1;
  ry=y;
    for(;;){
    if(rx<0 || bd[ry*8+rx]!=rc)
      break;
    rx--;
    if(rx>=0 && bd[ry*8+rx]==c)
      return true;
  }
// right
  rx=x+1;
  ry=y;
    for(;;){
    if(rx>=8 || bd[ry*8+rx]!=rc)
      break;
    rx++;
    if(rx<8 && bd[ry*8+rx]==c)
      return true;
  }
// up-left
  rx=x-1;
  ry=y-1;
    for(;;){
    if(rx<0 || ry<0 || bd[ry*8+rx]!=rc)
      break;
    rx--;
    ry--;
    if(rx>=0 && ry>=0 && bd[ry*8+rx]==c)
      return true;
  }
// down-left
  rx=x-1;
  ry=y+1;
    for(;;){
    if(rx<0 || ry>=8 || bd[ry*8+rx]!=rc)
      break;
    rx--;
    ry++;
    if(rx>=0 && ry<8 && bd[ry*8+rx]==c)
      return true;
  }
// up-right
  rx=x+1;
  ry=y-1;
    for(;;){
    if(rx>=8 || ry<0 || bd[ry*8+rx]!=rc)
      break;
    rx++;
    ry--;
    if(rx<8 && ry>=0 && bd[ry*8+rx]==c)
      return true;
  }
// down-right
  rx=x+1;
  ry=y+1;
    for(;;){
    if(rx>=8 || ry>=8 || bd[ry*8+rx]!=rc)
      break;
    rx++;
    ry++;
    if(rx<8 && ry<8 && bd[ry*8+rx]==c)
      return true;
  }
  return false;
}

static int put_num(char *bd,int c)
{
  int x,y,rx,ry,rc,num;
  int i;
  num=0;
  if(c==1) rc=2;
    else  rc=1;

  for(y=0;y<8;y++){
        for(x=0;x<8;x++){
          if(bd[y*8+x]!=0)
            continue;
// up
          rx=x;
          ry=y-1;
          for(;;){
            if(ry<0 || bd[ry*8+rx]!=rc)
              break;
            ry--;
            if(ry>=0 && bd[ry*8+rx]==c){
              num++;
                    goto loop_end;
            }
          }
// down
          rx=x;
          ry=y+1;
          for(;;){
            if(ry>=8 || bd[ry*8+rx]!=rc)
              break;
            ry++;
            if(ry<8 && bd[ry*8+rx]==c){
              num++;
              goto loop_end;
            }
          }
// left
          rx=x-1;
          ry=y;
          for(;;){
            if(rx<0 || bd[ry*8+rx]!=rc)
              break;
            rx--;
            if(rx>=0 && bd[ry*8+rx]==c){
              num++;
              goto loop_end;
            }
          }
// right
          rx=x+1;
          ry=y;
          for(;;){
            if(rx>=8 || bd[ry*8+rx]!=rc)
              break;
            rx++;
            if(rx<8 && bd[ry*8+rx]==c){
              num++;
              goto loop_end;
            }
          }
// up-left
          rx=x-1;
          ry=y-1;
          for(;;){
            if(rx<0 || ry<0 || bd[ry*8+rx]!=rc)
              break;
            rx--;
            ry--;
            if(rx>=0 && ry>=0 && bd[ry*8+rx]==c){
              num++;
              goto loop_end;
            }
          }
// down-left
          rx=x-1;
          ry=y+1;
          for(;;){
            if(rx<0 || ry>=8 || bd[ry*8+rx]!=rc)
              break;
            rx--;
            ry++;
            if(rx>=0 && ry<8 && bd[ry*8+rx]==c){
              num++;
              goto loop_end;
            }
          }
// up-right
          rx=x+1;
          ry=y-1;
          for(;;){
            if(rx>=8 || ry<0 || bd[ry*8+rx]!=rc)
              break;
            rx++;
            ry--;
            if(rx<8 && ry>=0 && bd[ry*8+rx]==c){
              num++;
              goto loop_end;
            }
          }
// down-right
          rx=x+1;
          ry=y+1;
          for(;;){
            if(rx>=8 || ry>=8 || bd[ry*8+rx]!=rc)
              break;
            rx++;
            ry++;
            if(rx<8 && ry<8 && bd[ry*8+rx]==c){
              num++;
              goto loop_end;
            }
          }
        loop_end:
    i=0;
        }
  }
  return(num);
}

static bool put_on_board(char *bd,int x,int y,int c)
{
  int rx,ry,rc;
  bool do_rev;
  do_rev=false;
  if(bd[y*8+x]!=0)
          return(do_rev);
  if(c==1) rc=2;
        else  rc=1;
// up
  rx=x;
  ry=y-1;
        for(;;){
    if(ry<0 || bd[ry*8+rx]!=rc)
      break;
    ry--;
    if(ry>=0 && bd[ry*8+rx]==c){
      rx=x;
      ry=y-1;
      do{
        bd[ry*8+rx]=c;
        ry--;
      }while(bd[ry*8+rx]!=c);
      do_rev=true;
                        break;
                }
  }
// down
  rx=x;
  ry=y+1;
        for(;;){
    if(ry>=8 || bd[ry*8+rx]!=rc)
      break;
    ry++;
    if(ry<8 && bd[ry*8+rx]==c){
      rx=x;
      ry=y+1;
      do{
        bd[ry*8+rx]=c;
        ry++;
      }while(bd[ry*8+rx]!=c);
      do_rev=true;
                        break;
                }
  }
// left
  rx=x-1;
  ry=y;
        for(;;){
    if(rx<0 || bd[ry*8+rx]!=rc)
      break;
    rx--;
    if(rx>=0 && bd[ry*8+rx]==c){
      rx=x-1;
      ry=y;
      do{
        bd[ry*8+rx]=c;
        rx--;
      }while(bd[ry*8+rx]!=c);
      do_rev=true;
                        break;
    }
  }
// right
  rx=x+1;
  ry=y;
        for(;;){
    if(rx>=8 || bd[ry*8+rx]!=rc)
      break;
    rx++;
    if(rx<8 && bd[ry*8+rx]==c){
      rx=x+1;
      ry=y;
      do{
        bd[ry*8+rx]=c;
        rx++;
      }while(bd[ry*8+rx]!=c);
      do_rev=true;
                        break;
                }
  }
// up-left
  rx=x-1;
  ry=y-1;
        for(;;){
    if(rx<0 || ry<0 || bd[ry*8+rx]!=rc)
      break;
    rx--;
    ry--;
    if(rx>=0 && ry>=0 && bd[ry*8+rx]==c){
      rx=x-1;
      ry=y-1;
      do{
        bd[ry*8+rx]=c;
        rx--;
        ry--;
      }while(bd[ry*8+rx]!=c);
      do_rev=true;
                        break;
                }
  }
// down-left
  rx=x-1;
  ry=y+1;
        for(;;){
    if(rx<0 || ry>=8 || bd[ry*8+rx]!=rc)
      break;
    rx--;
    ry++;
    if(rx>=0 && ry<8 && bd[ry*8+rx]==c){
      rx=x-1;
      ry=y+1;
      do{
        bd[ry*8+rx]=c;
        rx--;
        ry++;
      }while(bd[ry*8+rx]!=c);
      do_rev=true;
                        break;
                }
  }
// up-right
  rx=x+1;
  ry=y-1;
        for(;;){
    if(rx>=8 || ry<0 || bd[ry*8+rx]!=rc)
      break;
    rx++;
    ry--;
    if(rx<8 && ry>=0 && bd[ry*8+rx]==c){
      rx=x+1;
      ry=y-1;
      do{
        bd[ry*8+rx]=c;
        rx++;
        ry--;
      }while(bd[ry*8+rx]!=c);
      do_rev=true;
                        break;
                }
  }
// down-right
  rx=x+1;
  ry=y+1;
        for(;;){
    if(rx>=8 || ry>=8 || bd[ry*8+rx]!=rc)
      break;
    rx++;
    ry++;
    if(rx<8 && ry<8 && bd[ry*8+rx]==c){
      rx=x+1;
      ry=y+1;
      do{
        bd[ry*8+rx]=c;
        rx++;
        ry++;
      }while(bd[ry*8+rx]!=c);
      do_rev=true;
                        break;
                }
  }
  if(do_rev==true)
    bd[y*8+x]=c;
  return(do_rev);
}

static void think(char *bd,int *x,int *y,int c,int level)
{
  int rx,ry,p,maxx,maxy,maxp;

  maxp=-1000;
  for(ry=0;ry<8;ry++){
    for(rx=0;rx<8;rx++){
      if (can_put(bd, rx, ry, c)==false)
        continue;
      if (komasuu>=56 && level>=4)
        p=rev_point(bd, rx, ry, c, 8);
      else if (komasuu>=58 && level==3)
        p=rev_point(bd, rx, ry, c, 6);
      else if (komasuu>=60)
        p=rev_point(bd, rx, ry, c, 4);
      else
        p=rev_point(bd, rx, ry, c, level);
      if(p>maxp){
        maxx=rx;
        maxy=ry;
        maxp=p;
      }
    }
  }

  *x=maxx;
  *y=maxy;
  return;
}

static const int boardp[8][8]={
  32, -4, 4, 4, 4, 4, -4,32,
  -4,-16, 1, 1, 1, 1,-16,-4,
   4,  1, 1, 1, 1, 1,  1, 4,
   4,  1, 1, 1, 1, 1,  1, 4,
   4,  1, 1, 1, 1, 1,  1, 4,
   4,  1, 1, 1, 1, 1,  1, 4,
  -4,-16, 1, 1, 1, 1,-16,-4,
  32, -4, 4, 4, 4, 4, -4,32};

static int rev_point(char *bd,int x,int y,int c,int level)
{
  char bdc[8][8];
  int num;
  int point;
  int rx, ry, rc, maxp, po;

  if(c==1) rc=2;
  else rc=1;

  if (put_num(bd, c)==0 && put_num(bd, rc)==0){
    num=0;
    for(ry=0;ry<8;ry++){
      for(rx=0;rx<8;rx++){
        if (bd[ry*8+rx]==c)
          num++;
        else if (bd[ry*8+rx]==rc)
          num--;
      }
    }
    return(num*8);
  }

  for(ry=0;ry<8;ry++)
    for(rx=0;rx<8;rx++)
      bdc[ry][rx]=bd[ry*8+rx];

  if(put_on_board((char *)bdc, x, y, c)==false)
    return(-500);

  if (level<=1){
    num=0;
    point=0;
    for(ry=0;ry<8;ry++){
      for(rx=0;rx<8;rx++){
        if (bdc[ry][rx]==c){
          num++;
          point=point+boardp[ry][rx];
        }else if(bdc[ry][rx]==rc){
          num--;
          point=point-boardp[ry][rx];
        }
      }
    }

    if(komasuu<30)
      po=((4-put_num((char *)bdc, rc))*4+point)*4+random(1, 8);
    else if(komasuu<56)
      po=((4-put_num((char *)bdc, rc))+point)*4+random(1, 8);
    else
      po=num*8;

    return(po);
  }

  maxp=-1000;
  for(ry=0;ry<8;ry++){
    for(rx=0;rx<8;rx++){
      po=rev_point((char *)bdc, rx, ry, rc, level-1);
      if (po>maxp)
        maxp=po;
    }
  }
  if (maxp==-1000)
    maxp=0;
  return(-maxp);
}

void tone (uint8_t pin, uint16_t freq, uint16_t leng)
{
  if (tone_init==0){
    ledcSetup(1, 12000, 8);
    ledcAttachPin(pin, 1);
    tone_init=1;
  }
  ledcWriteTone(1, freq);
}

void noTone (uint8_t pin)
{
  ledcWriteTone(1, 1);
}
